#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : chatwhoosh
# @Time         : 2023/4/26 19:04
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :
import pandas as pd

from meutils.pipe import *
from meutils.hash_utils import md5
from meutils.easy_search.es import EasySearch

from chatllm.applications import ChatBase

from whoosh.fields import *
from jieba.analyse import ChineseAnalyzer


class ChatWhoosh(ChatBase, EasySearch):

    def __init__(self, indexdir='whoosh_index', indexname='MAIN', **kwargs):
        ChatBase.__init__(self, **kwargs)
        EasySearch.__init__(self, indexdir, indexname)
        self.recall = pd.DataFrame({'id': [], 'text': [], 'score': []})

    def qa(self, query, topk=3, threshold=0.66, **kwargs):
        df = self.find(query, topk, threshold)
        if len(df) == 0:
            logger.warning('召回内容为空!!!')
        knowledge_base = '\n'.join(df.text)

        return self._qa(query, knowledge_base, **kwargs)

    def find(self, query, topk=3, threshold=0.66, **kwargs):
        df = super().find(defaultfield='text', querystring=query, limit=topk, **kwargs)
        if len(df):
            self.recall = df.query(f'score > {threshold}')
        return self.recall

    def create_index(self, texts, id_mapping=md5, **kwargs):
        ids = map(id_mapping, texts)
        df = pd.DataFrame({'id': ids, 'text': texts})
        schema = Schema(
            id=ID(stored=True),
            text=TEXT(stored=True, analyzer=ChineseAnalyzer(cachesize=-1))  # 无界缓存加速
        )
        super().create_index(df, schema, **kwargs)


if __name__ == '__main__':
    from chatllm.applications.chatwhoosh import ChatWhoosh

    cw = ChatWhoosh(indexdir='whoosh_index')
    cw.create_index(texts=['周杰伦'] * 10)
    print(cw.find('周杰伦'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : crawler4qa
# @Time         : 2023/4/24 18:17
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :


from meutils.pipe import *
from meutils.request_utils.crawler import Crawler

from chatllm.applications import ChatBase


class Crawler4QA(ChatBase):

    def __init__(self, chat_func):
        super().__init__(chat_func)

    def qa(self, query,
           url="https://top.baidu.com/board?tab=realtime",
           xpath='//*[@id="sanRoot"]/main/div[2]/div/div[2]/div[*]/div[2]/a/div[1]//text()', **kwargs):
        knowledge_base = Crawler(url).xpath(xpath)  # 爬虫获取知识库

        return self._qa(query, knowledge_base, **kwargs)


if __name__ == '__main__':
    from chatllm.utils import load_llm4chat

    chat_func = load_llm4chat(
        model_name_or_path="/Users/betterme/PycharmProjects/AI/CHAT_MODEL/chatglm",
        device='mps',
    )

    qa = Crawler4QA(chat_func=chat_func)

    for i, _ in qa(query='提取人名'):
        print(i, flush=True)
        sys.stdout.flush()

    pprint(qa.knowledge_base)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : utils
# @Time         : 2023/4/20 12:50
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :
import torch
from transformers import AutoTokenizer, AutoModel

from meutils.pipe import *
from chatllm.utils.gpu_utils import load_chatglm_on_gpus

DEVICE = (
    os.environ['DEVICE'] if 'DEVICE' in os.environ
    else "cuda" if torch.cuda.is_available()
    else "mps" if torch.backends.mps.is_available()
    else "cpu"
)
if LOCAL_HOST.startswith('10.219'):
    MODEL_PATH = "/Users/betterme/PycharmProjects/AI/CHAT_MODEL/chatglm"


# todo 多卡 https://github.com/THUDM/ChatGLM-6B#%E5%A4%9A%E5%8D%A1%E9%83%A8%E7%BD%B2

def textsplitter(text, chunk_size=512, overlap_rate=0.2, sep=''):  # 简单粗暴
    return text.lower().split() | xjoin(sep) | xgroup(chunk_size, overlap_rate)


def load_llm(model_name_or_path="THUDM/chatglm-6b", device=DEVICE, num_gpus=2, **kwargs):
    model = AutoModel.from_pretrained(model_name_or_path, trust_remote_code=True)
    tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, trust_remote_code=True)

    if torch.cuda.is_available() and device.lower().startswith("cuda"):
        num_gpus = min(1, num_gpus, torch.cuda.device_count())

        if num_gpus == 1:  # 单卡
            model = model.half().cuda()
            # model.transformer.prefix_encoder.float()
        elif 'chatglm' in model_name_or_path:  # chatglm多卡
            model = load_chatglm_on_gpus(model_name_or_path, num_gpus)

    else:
        model = model.float().to(device)

    return model.eval(), tokenizer


def load_llm4chat(model_name_or_path="THUDM/chatglm-6b", device=DEVICE, num_gpus=1, stream=True, **kwargs):
    model, tokenizer = load_llm(model_name_or_path, device, num_gpus, **kwargs)
    if stream and hasattr(model, 'stream_chat'):
        return partial(model.stream_chat, tokenizer=tokenizer)  # 可以在每一次生成清GPU
    else:
        return partial(model.chat, tokenizer=tokenizer)


if __name__ == '__main__':
    model, tokenizer = load_llm("/CHAT_MODEL/chatglm", device='cpu')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : ChatPDF
# @Time         : 2023/4/21 11:44
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from meutils.office_automation.pdf import extract_text, pdf2text

from chatllm.utils import textsplitter
from chatllm.applications.chatann import ChatANN


class ChatPDF(ChatANN):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def create_index(self, file_or_text, textsplitter=textsplitter):  # todo 多篇 增加 parser loader

        texts = extract_text(file_or_text)
        texts = textsplitter(texts)
        return super().create_index(texts)


if __name__ == '__main__':
    # filename = '../../data/财报.pdf'
    # bytes_array = Path(filename).read_bytes()
    # texts = extract_text(bytes_array)
    # texts = textsplitter(texts)
    # print(texts)
    from chatllm.applications.chatpdf import ChatPDF

    qa = ChatPDF(encode_model='nghuyong/ernie-3.0-nano-zh')  # 自动建索引
    qa.create_index('../../data/财报.pdf')
    qa.load_llm4chat(model_name_or_path='/Users/betterme/PycharmProjects/AI/CHAT_MODEL/chatglm', device='cpu')

    list(qa(query='东北证券主营业务', topk=1, threshold=0.8))

    # 召回结果
    print(qa.recall)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : qa
# @Time         : 2023/4/20 17:47
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from langchain.chains import RetrievalQA
from langchain.prompts.prompt import PromptTemplate
from langchain.vectorstores import FAISS

# ME
from chatllm._his._chatllm import ChatLLM

RetrievalQA.return_source_documents = True


class QA(object):
    def __init__(self, chatllm: ChatLLM, faiss_ann: FAISS = None, document_prompt: PromptTemplate = None):
        """

        :param chatllm:
        """
        self.chatllm = chatllm
        self.faiss_ann = faiss_ann
        self.document_prompt = document_prompt if document_prompt else self.default_document_prompt

    @property
    def default_document_prompt(self) -> PromptTemplate:
        prompt_template = """
        基于以下已知信息，简洁和专业的来回答用户的问题。
        如果无法从中得到答案，请说 "根据已知信息无法回答该问题" 或 "没有提供足够的相关信息"，不允许在答案中添加编造成分，答案请使用中文。
        已知内容:
        {context}
        问题:
        {question}
        """.strip()

        return PromptTemplate(template=prompt_template, input_variables=["context", "question"])

    def get_knowledge_based_answer(self, query, max_turns=3, top_k=4, **kwargs):
        assert self.faiss_ann

        # 设置chatllm参数，# history会被储存？
        self.chatllm.set_chat_kwargs(**kwargs)
        self.chatllm.max_turns = max_turns

        llm_chain = RetrievalQA.from_llm(
            llm=self.chatllm,
            retriever=self.faiss_ann.as_retriever(search_kwargs={"k": top_k}),  # todo: 重复实例化优化
            prompt=self.document_prompt
        )
        # llm_chain.combine_documents_chain.document_prompt = PromptTemplate(
        #     input_variables=["page_content"], template="{page_content}"
        # )
        # 官方默认，要不要覆盖
        # document_prompt = PromptTemplate(
        #     input_variables=["page_content"], template="Context:\n{page_content}"
        # )

        result = llm_chain({"query": query})
        return result

    def get_llm_answer(self, query, max_turns=3, **kwargs):  # 重复代码
        self.chatllm.set_chat_kwargs(**kwargs)
        self.chatllm.max_turns = max_turns

        return self.chatllm._call(query)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : stream_api
# @Time         : 2023/4/28 09:04
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :
import json

from meutils.pipe import *
from meutils.str_utils import json_loads

from fastapi.responses import Response, StreamingResponse
from fastapi import FastAPI, Form, Depends, File, UploadFile, Body, Request, BackgroundTasks


class ChatApi(FastAPI):

    def __init__(self, title: str = "ChatLLM API", **kwargs):
        super().__init__(title=title, **kwargs)
        self.handler_func = lambda query='自定义 handler_func': query

    def run(self):
        import uvicorn
        uvicorn.run(self)

    def register(self, path='/', handler_func=None, methods=None):
        self.handler_func = handler_func or self.handler_func
        self.api_route(path=path, methods=methods)(self.handler)

    async def handler(self, request: Request):
        input = request.query_params._dict
        body = await request.body()

        if body.startswith(b'{'):  # 主要分支 # json={}
            input.update(json_loads(body))  # json_loads

        query = input.get('query', '')

        if inspect.isgeneratorfunction(self.handler_func):
            return StreamingResponse(self.handler_func(query), media_type='text/event-stream')

        return {'data': self.handler_func(query)}


if __name__ == '__main__':

    app = ChatApi()


    def gen_data(query):
        for i in range(5):
            time.sleep(i)
            yield f"{query} {i}\n"
    def gen_data_(query):
        for i in range(10):
            time.sleep(i)
            yield f"{query} {i}\n"


    app.register('', handler_func=gen_data)
    app.register('/s', handler_func=gen_data_)

    app.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : chatpdf
# @Time         : 2023/4/25 17:01
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from chatllm.applications import ChatBase
from chatllm.utils import load_llm4chat

import streamlit as st
from appzoo.streamlit_app.utils import display_pdf, reply4input

st.set_page_config('🔥ChatLLM', layout='centered', initial_sidebar_state='collapsed')


@st.cache_resource
def get_chat_func():
    chat_func = load_llm4chat(
        model_name_or_path="/Users/betterme/PycharmProjects/AI/CHAT_MODEL/chatglm"
    )
    return chat_func


chat_func = get_chat_func()

qa = ChatBase(chat_func=chat_func)


def reply_func(query):
    for response, _ in qa(query=query):
        yield response


# def reply_func(x):
#     for i in range(10):
#         time.sleep(1)
#         x += str(i)
#         yield x


container = st.container()  # 占位符
text = st.text_area(label="用户输入", height=100, placeholder="请在这儿输入您的问题")
# knowledge_base = st.sidebar.text_area(label="知识库", height=100, placeholder="请在这儿输入您的问题")

if st.button("发送", key="predict"):
    with st.spinner("AI正在思考，请稍等........"):
        history = st.session_state.get('state')
        st.session_state["state"] = reply4input(text, history, container=container, reply_func=reply_func)
